@@ -8,6 +8,8 @@ from .ninjax import Variable
 from .ninjax import cond
 from .ninjax import scan
 from .ninjax import grad
+from .ninjax import jit
+from .ninjax import pmap
 from .ninjax import HaikuModule
 from .ninjax import FlaxModule
 from .ninjax import OptaxModule
